/*
 * encode.c - write target code for pcc3 compiler
 *
 * The target is the IA-32 Intel Architecture.
 *
 * This code also assumes the compiler will run on an IA-32 machine.
 *
 * All operands are left on top of the run-time stack so no register
 * allocation is performed. Registers are only used as temporaries
 * while doing individual operations.
 */

#include "defs.h"
#include "types.h"
#include "tree.h"
#include "symtab.h"
#include "encode.h"
#include "globals.h"
#include "message.h"
#include "semUtils.h"

static void encodeStructUnionAssign(BinopTree b);
static void encodeAssign(BinopTree b);
static void encodeFcall(FcallTree b);
static void encodeReturn(ReturnTree r);

static void encodeConvert(UnopTree u);

/*
 *  Track size of the locals area for current function. Local variables
 *  are allocated in the locals area from the low addresses back to the
 *  high addresses. That is, first local variable is at the lowest address.
 *  This may appear backwards as the variables asigned the smaller addresses
 *  will appear at the higher negative offsets in the generated code.
 */
static int sizeOfLocalsArea;

/*
 *  genLabel: generate a unique label number
 *            prepend L$ to get full label name
 *            e.g.: L$5
 */
unsigned int genLabel(void)
{
  static unsigned int nextLabel = 0;

  return nextLabel++;
}

/********************************************
  encode_init

  output output file prologue
********************************************/
void encode_init(void)      /* beginning of program */
{
  emit("\t.data");          /* global variables will be first */
}


/************************************************************
  encode_global_var_decl
*************************************************************/

void encode_global_var_decl(ST_ID id, Type type)
{
  int size;
       
  size = computeSize( type );
  emit("\t.comm\t%s, %d", st_get_id_str(id), size);
}


/************************************************************
  encode_static_var_decl
*************************************************************/

void encode_static_var_decl(ST_ID id, Type type)
{
  int size;
        
  size = computeSize( type );
  emit("\t.lcomm\t%s, %d", st_get_id_str(id), size);
}

/* function to generate code for local static variables
 * defferent from global static for the following case:
 * fun1() { static int a; }
 * fun2() { static float a; }
 */
void encode_local_static( ST_ID id, unsigned int label, Type type )
{
    int size;

    size = computeSize( type );
    emit("\t.data");
    emit("\t.lcomm\t%s$L%d, %d", st_get_id_str(id), label, size );
    emit("\t.text");
}

/* compute the size in bytes for the given Type */
int computeSize( Type type )
{
  switch(typeQuery(type)) {

  case TYFLOAT :
    return 4;

  case TYDOUBLE :
    return 8;

  case TYLONGDOUBLE :
    return 12; /* but only 10 bytes actually used? */

  case TYUNSIGNEDINT :
    return 4;

  case TYUNSIGNEDCHAR :
    return 1;

  case TYUNSIGNEDSHORTINT :
    return 2;

  case TYUNSIGNEDLONGINT :
    return 4;

  case TYSIGNEDCHAR :
    return 1;

  case TYSIGNEDINT :
    return 4;

  case TYSIGNEDLONGINT:
    return 4;

  case TYSIGNEDSHORTINT:
    return 2;

  case TYVOID:
    bug("TYVOID in computeSize()"); 
    return 0; /*NOTREACHED*/

  case TYPOINTER:        /* all pointers are of the same size */
    return 4;

  case TYARRAY:
  {
    DimFlag dimflag;
    unsigned int dim;
    Type elementType;

    elementType = typeQueryArray(type, &dimflag, &dim);
    if ( dimflag == DIMENSION_PRESENT )
    {
      return computeSize(elementType) * dim;
    }
    else
    {
      bug("array with no dimension in computeSize()");
      return 0; /*NOTREACHED*/
    }
  }/* end of case array */

  case TYERROR:
    return 0;

  case TYSTRUCT:
    return computeStructSize(type);

  case TYUNION:
    return computeUnionSize(type);

  default: 
    bug("unsupported type in computeSize()\n");
    return 0; /*NOTREACHED*/

  }/* end of switch */
  return 0; /*NOTREACHED*/

}/* end of computeSize */


/* compute the alignment of the given Type */
int determineAlignment( Type type )
{
  switch( typeQuery(type) ) {
  case TYFLOAT :
  case TYDOUBLE :
  case TYLONGDOUBLE :
    return 4;
  case TYUNSIGNEDCHAR :
  case TYSIGNEDCHAR :
    return 1;
  case TYUNSIGNEDSHORTINT :
  case TYSIGNEDSHORTINT:
    return 2;
  case TYUNSIGNEDLONGINT :
  case TYUNSIGNEDINT :
  case TYSIGNEDINT :
  case TYSIGNEDLONGINT:
  case TYPOINTER:
    return 4;

  case TYVOID:
    bug("TYVOID in determineAlignment()"); 
    return 0; /*NOTREACHED*/

  case TYERROR:
    return 0;

  case TYARRAY:    /* equal to its element type alignment */
  {
    DimFlag dimflag;
    unsigned int dim;
    Type elementType;

    elementType = typeQueryArray(type, &dimflag, &dim);
    return determineAlignment(elementType);
  }

  case TYSTRUCT:
  case TYUNION:
    return 4;      /* worst case alignment */

  default:
    bug("unsupported type in determineAlignment()\n");      
    return 0; /*NOTREACHED*/

}/* end of switch */

}/* end of determineAlignment */

int worstCaseAlignment(void)
{
  return 4;
}

/* is the given Type derer-able on the target architecture? */
BOOLEAN isDerefable(Type t)
{
  switch(typeQuery(t)) {

  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  case TYPOINTER:
    return TRUE;

  case TYSTRUCT:
  case TYUNION:
  case TYARRAY:
  case TYERROR:
  case TYVOID:
    return FALSE;

  case TYFUNCTION:
    return FALSE;

  default:
    bug("unknown type in isDerefable");
  }

  return FALSE; /*NOTREACHED*/
}


/*
 * determineIA32Type: Type -> IA-32 single char type spec (ie b, w, l, s, etc)
 *
 */
char determineIA32Type(Type t)
{
  switch (typeQuery(t)) {
  case TYFLOAT:
    return 's';
  case TYDOUBLE:
    return 'l';
  case TYLONGDOUBLE:
    return 't';
  case TYSIGNEDLONGINT:
    return 'l';
  case TYSIGNEDSHORTINT:
    return 'w';
  case TYSIGNEDINT:
    return 'l';
  case TYUNSIGNEDLONGINT:
    return 'l';
  case TYUNSIGNEDSHORTINT:
    return 'w';
  case TYUNSIGNEDINT:
    return 'l';
  case TYUNSIGNEDCHAR:
    return 'b';
  case TYSIGNEDCHAR:
    return 'b';
  case TYPOINTER:
    return 'l';
  case TYVOID:
    bug("void type seen in determineIA32Type()");
    break;
  case TYERROR:
    bug("error type seen in determineIA32Type()");
    break;
  case TYSTRUCT:
    bug("struct type seen in determineIA32Type()");
    break;
  case TYUNION:
    bug("union type seen in determineIA32Type()");
    break;
  case TYARRAY:
    bug("array type seen in determineIA32Type()");
    break;
  case TYFUNCTION:
    bug("function type seen in determineIA32Type()");
    break;
  default:
    bug("unknown type seen in determineIA32Type()");
  }
  return ' '; /*NOTREACHED*/
}


/* 
 *  encodeAssignDetail: make details of encoding ASSIGN generally
 *                     available. Used, for example, in compound
 *                     assignment operators. This only supports
 *                     the "basic" types (ie not struct/union).
 *
 *  type - Type of value to be assigned.
 *
 *  t - Tree being encoded.
 *
 */
void encodeAssignDetail(Type type, Tree t)
{
  switch(typeQuery(type)) {

  case TYFLOAT:

    /* pop rhs (32 bits) */
    emit("\tpopl\t%%eax"); 

    /* pop lhs */
    emit("\tpopl\t%%edx");

    /* actually do the assign */
    emit("\tmovl\t%%eax, (%%edx)");

    /* push result (32 bits) */
    emit("\tpushl\t%%eax");

    break;

  case TYDOUBLE:

    /* get rhs (64 bits) from top of stack */
    emit("\tfldl\t(%%esp)");

    /* pop */
    emit("\taddl\t$8, %%esp");

    /* pop lhs */
    emit("\tpopl\t%%eax");

    /* actually do the assign */
    emit("\tfstl\t(%%eax)");

    /* alloc spot (64 bits) for result */
    emit("\tsubl\t$8, %%esp");

    /* put result on top of stack (and pop fp register stack) */
    emit("\tfstpl\t(%%esp)");

    break;

  case TYLONGDOUBLE:

    /* get rhs (80 bits stored in 12 bytes) from top of stack */
    emit("\tfldt\t(%%esp)");

    /* pop */
    emit("\taddl\t$12, %%esp");

    /* pop lhs */
    emit("\tpopl\t%%eax");

    /* actually do the assign (also pops the fp register stack) */
    emit("\tfstpt\t(%%eax)");

    /* re-load the result */
    emit("\tfldt\t(%%eax)");

    /* alloc spot (80 bits stored in 12 bytes) for result */
    emit("\tsubl\t$12, %%esp");

    /* put result on top of stack (and pop fp register stack) */
    emit("\tfstpt\t(%%esp)");

    break;

  case TYSIGNEDCHAR:
  case TYUNSIGNEDCHAR:

    /* pop rhs */
    emit("\tpopl\t%%eax");

    /* pop lhs */
    emit("\tpopl\t%%edx");

    /* do the assign */
    emit("\tmovb\t%%al, (%%edx)");

    /* push the result */
    emit("\tpushl\t%%eax");
    break;

  case TYSIGNEDSHORTINT:
  case TYUNSIGNEDSHORTINT:

    /* pop rhs */
    emit("\tpopl\t%%eax");

    /* pop lhs */
    emit("\tpopl\t%%edx");

    /* do the assign */
    emit("\tmovw\t%%ax, (%%edx)");

    /* push the result */
    emit("\tpushl\t%%eax");
    break;

  case TYSIGNEDINT:
  case TYUNSIGNEDINT:
  case TYSIGNEDLONGINT:
  case TYUNSIGNEDLONGINT:
  case TYPOINTER:

    /* pop rhs */
    emit("\tpopl\t%%eax");

    /* pop lhs */
    emit("\tpopl\t%%edx");

    /* do the assign */
    emit("\tmovl\t%%eax, (%%edx)");

    /* push the result */
    emit("\tpushl\t%%eax");
    break;

  case TYSTRUCT:
  case TYUNION:
  case TYARRAY:
  case TYERROR:
  case TYVOID:
  case TYFUNCTION:
  default:
    bugT(t, "unknown type in encodeAssignDetail");
  }

  /* mark the tree as visited by the code generator */
  ((ExpTree) t)->reg = setReg();
}

/* 
 *  encodeDerefDetail: make details of encoding DEREF generally
 *                     available. Used, for example, in compound
 *                     assignment operators.
 *
 *  type - Type of value to be loaded.
 *
 *  t - Tree being encoded.
 *
 */
void encodeDerefDetail(Type type, Tree t)
{
  /* pop address to be deref-ed */
  emit("\tpopl\t%%eax");

  switch(typeQuery(type)) {

  case TYFLOAT:

    /* load to fp register */
    emit("\tflds\t(%%eax)");

    /* alloc spot (32 bits) for result */
    emit("\tsubl\t$4, %%esp");

    /* put value on top of stack (and pop fp register stack) */
    emit("\tfstps\t(%%esp)");

    break;

  case TYDOUBLE:

    /* load to fp register */
    emit("\tfldl\t(%%eax)");

    /* alloc spot (64 bits) for result */
    emit("\tsubl\t$8, %%esp");

    /* put value on top of stack (and pop fp register stack) */
    emit("\tfstpl\t(%%esp)");

    break;

  case TYLONGDOUBLE:

    /* load to fp register */
    emit("\tfldt\t(%%eax)");

    /* alloc spot (80 bits stored in 12 bytes) for result */
    emit("\tsubl\t$12, %%esp");

    /* put value on top of stack (and pop fp register stack) */
    emit("\tfstpt\t(%%esp)");

    break;

  case TYSIGNEDSHORTINT:

    /* do the deref */
    emit("\tmovswl\t(%%eax), %%eax");

    /* push result onto stack */
    emit("\tpushl\t%%eax");
    break;

  case TYUNSIGNEDSHORTINT:

    /* do the deref */
    emit("\tmovzwl\t(%%eax), %%eax");

    /* push result onto stack */
    emit("\tpushl\t%%eax");
    break;

  case TYUNSIGNEDCHAR:

    /* do the deref */
    emit("\tmovzbl\t(%%eax), %%eax");

    /* push result onto stack */
    emit("\tpushl\t%%eax");
    break;

  case TYSIGNEDCHAR:

    /* do the deref */
    emit("\tmovsbl\t(%%eax), %%eax");

    /* push result onto stack */
    emit("\tpushl\t%%eax");
    break;

  case TYSIGNEDLONGINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDINT:
  case TYPOINTER:

    /* do the deref */
    emit("\tmovl\t(%%eax), %%eax");

    /* push result onto stack */
    emit("\tpushl\t%%eax");
    break;

  case TYSTRUCT:
  case TYUNION:
  case TYARRAY:
  case TYERROR:
  case TYVOID:
  case TYFUNCTION:
    bugT(t, "unexpected type in encodeDerefDetail");

  default:
    bugT(t, "unknown type in encodeDerefDetail");
  }

  /* mark the tree as visited by the code generator */
  ((ExpTree) t)->reg = setReg();
}

/*
 *  encodeDup
 *
 *  A DUP node indicates that a subtree has two parents. The DUP
 *  node is placed on top of the subtree on the path that will be
 *  evaluated first. Therefore, to encode DUP, we must simply duplicate
 *  the top of stack.
 *
 *  Currently this routine is only used to implement compound assignments.
 *  Therefore the only value DUP-ed is an address value. So this routine
 *  will only DUP a 4-byte value.
 */
static void encodeDup(UnopTree u)
{
  /* copy and push the value */
  emit("\tmovl\t(%%esp), %%eax");
  emit("\tpushl\t%%eax");

  /* mark the tree as visited by the code generator */
  u->expn.reg = setReg();

  return;
}

/*
 * encodeUnop: code generation for unary operators
 *
 * Note: No support for SIZEOF as it will be fully handled at compile time
 *       (during semantic analysis).
 */
static void encodeUnop(UnopTree u)
{
  switch (u->op) {
  case NO_OP:
    break;
  case DEREF_OP:
    encodeDerefDetail(u->expn.type, (Tree) u);
    break;
  case DUP_OP:
    encodeDup(u);
    break;
  case INCRA_OP:  
    encodeIncra(u);
    break;
  case DECRA_OP:
    encodeDecra(u);
    break;
  case INCRB_OP:  
    encodeIncrb(u);
    break;
  case DECRB_OP:
    encodeDecrb(u);
    break;
  case USUB_OP:
    encodeUsub(u);
    break;
  case LNOT_OP:
    encodeLnot(u);
    break;
  case CONVERT_OP:
    encodeConvert(u);
    break;
  case SIZEOF_OP:
    bugT((Tree) u, "SIZEOF not expected in encodeUnary");
    break;
  case ADDR_OF_OP:
    encodeAddrOf(u);
    break;
  case COMPLMT_OP:
    encodeComplmt(u);
    break;
  case UADD_OP:
    bugT((Tree) u, "UADD seen in encodeUnary!");
    break;
  default:
    bugT((Tree) u, "unknown operation in encodeUnary");
  }
}

/*
 * encodeBinop: code generation for binary operators
 *
 */
static void encodeBinop(BinopTree b)
{
  switch (b->op) {

  case INDEX_OP:
    encodeIndex(b);
    break;

  case LSHIFT_OP:
    encodeLshift(b);
    break;

  case RSHIFT_OP:
    encodeRshift(b);
    break;

  case ANDD_OP:
    encodeAndd(b);
    break;

  case XORR_OP:
    encodeXorr(b);
    break;

  case ORR_OP:
    encodeOrr(b);
    break;

  case MULT_ASSIGN_OP:
    encodeMultAssign(b);
    break;

  case DIV_ASSIGN_OP:
    encodeDivAssign(b);
    break;

  case MOD_ASSIGN_OP:
    encodeModAssign(b);
    break;

  case ADD_ASSIGN_OP:
    encodeAddAssign(b);
    break;

  case SUB_ASSIGN_OP:
    encodeSubAssign(b);
    break;

  case LEFT_ASSIGN_OP:
    encodeLeftAssign(b);
    break;

  case RIGHT_ASSIGN_OP:
    encodeRightAssign(b);
    break;

  case AND_ASSIGN_OP:
    encodeAndAssign(b);
    break;

  case XOR_ASSIGN_OP:
    encodeXorAssign(b);
    break;

  case OR_ASSIGN_OP:
    encodeOrAssign(b);
    break;

  case MULT_OP:
    encodeMult(b);
    break;

  case DIV_OP:
    encodeDiv(b);
    break;

  case MOD_OP:
    encodeMod(b);
    break;

  case ADD_OP:
    encodeAdd(b);
    break;

  case SUB_OP:
    encodeSub(b);
    break;

  case LT_OP:
    encodeLt(b);
    break;

  case GT_OP:
    encodeGt(b);
    break;

  case LTE_OP:
    encodeLte(b);
    break;

  case GTE_OP:
    encodeGte(b);
    break;

  case EQL_OP:
    encodeEql(b);
    break;

  case NEQL_OP:
    encodeNeql(b);
    break;

  case ASSIGN_OP:
    encodeAssign(b);
    break;

  case LOR_OP:
  case LAND_OP:
  case COMMA_OP:
    bugT((Tree) b, "unexpected LOR, LAND or COMMA seen in encodeBinary");

  default:
    bugT((Tree) b, "unknown operation in encodeBinary");
  }
}

/*
 *  encode: traverses an AST and generates code in a (basically)
 *          bottom-up fashion.
 */

void encode( Tree t )
 {
   UnopTree u;
   BinopTree b;
   TriopTree tr;
   FcallTree f;
   EseqTree e;
   PtrTree ptr;
   FieldRefTree fr;
   CastopTree c;
   VarTree v;
   IntConstTree i;
   FpConstTree fp;
   StringLitTree s;
   LabelTree la;
   CaseTree ca;
   DefaultTree d;
   SeqTree se;
   IfElseTree ie;
   SwitchTree sw;
   WhileTree w;
   DoWhileTree dw;
   ForTree fo;
   GotoTree g;
   ContinueTree co;
   BreakTree br;
   ReturnTree r;
   int labNum;

   if ( t == NULL )
   {
      return;
   }

   switch( t->tag ){

      case UNOP_TAG:
               u = ( UnopTree ) t;
               if (typeQuery(u->expn.type) == TYERROR) break;
               if (u->expn.reg) break;
               encode( ( Tree ) u->left );
               encodeUnop(u);
               break;

      case BINOP_TAG:
               b = ( BinopTree ) t;

               if (typeQuery(b->expn.type) == TYERROR) break;
               if (b->expn.reg) break;

               /* logical operators cannot do default evaluation order */
               if (b->op == LAND_OP)
               {
                 encodeLand(b);
               }
               else if (b->op == LOR_OP)
               {
                 encodeLor(b);
               }
               /* convenient to override default evaluation for comma, too */
               else if (b->op == COMMA_OP)
               {
                 encodeComma(b);
               }
               else
               {
                 encode( ( Tree ) b->left );
                 encode( ( Tree ) b->right );
                 encodeBinop(b);
               }
               break;

      case TRIOP_TAG:
               tr = ( TriopTree ) t;
               if (typeQuery(tr->expn.type) == TYERROR) break;
               if (tr->expn.reg) break;
               if (tr->op == COND_OP)
               {
                 encodeCond(tr);
               }
               else
               {
                 bugT(t, "encode: unknown triop");
               }
               break;

      case FCALL_TAG:
               f = ( FcallTree ) t;
               if (typeQuery(f->expn.type) == TYERROR) break;
               if (f->expn.reg) break;
               encodeFcall(f);
               break;

      case ESEQ_TAG:
               e = ( EseqTree ) t;
               if (e->expn.reg) break;
               encode( ( Tree ) e->left );
               encode( ( Tree ) e->right );
               break;

      case PTR_TAG:
               ptr = ( PtrTree ) t;
               if (typeQuery(ptr->expn.type) == TYERROR) break;
               if (ptr->expn.reg) break;
               encode( ( Tree ) ptr->left );
               encodePtr( ptr );
               break;

      case FIELDREF_TAG:
               fr = ( FieldRefTree ) t;
               if (typeQuery(fr->expn.type) == TYERROR) break;
               if (fr->expn.reg) break;
               encode( ( Tree ) fr->left );
               encodeFieldRef( fr );
               break;

      case CASTOP_TAG:
               c = ( CastopTree ) t;
               if (typeQuery(c->expn.type) == TYERROR) break;
               if (c->expn.reg) break;
               encode( ( Tree ) c->left );
               encodeCast(c);
               break;

      case VAR_TAG:
               v = ( VarTree ) t;
               if (typeQuery(v->expn.type) == TYERROR) break;
               if (v->expn.reg) break;

               /* mark tree as visited */
               v->expn.reg = setReg();

               /* global variable, local extern, or function */
               if ((v->block == 0) || (v->class == EXTERN_SC) ||
                   (typeQuery(v->expn.type) == TYFUNCTION))
               {
                 emit("\tpushl\t$%s", st_get_id_str(v->name));
               }
               /* local variable or parameter */
               else
               {
                 /* for static locals use a munged name */
                 if (v->class == STATIC_SC)
                 {
                   emit("\tpushl\t$%s$L%u", st_get_id_str(v->name),
                     v->staticLabel);
                 }
                 /*
                  *  local/parameter is on the stack
                  *
                  *  Parameters are at positive offsets from the ebp
                  *  while locals are at negative offsets from the ebp.
                  *
                  *  Add 8 to the parameter offset in order to skip over
                  *  the old ebp and the old eip.
                  *
                  *  Locals are assigned spots starting at the bottom of
                  *  of the locals area and working back toward 0(%ebp).
                  */
                 else
                 {
                   switch (v->stdrTag) {

                   case PDECL:
                     emit("\tleal\t%d(%%ebp), %%eax", v->offset + 8);
                     break;

                   case LDECL:
                     emit("\tleal\t%d(%%ebp), %%eax",
                       -(sizeOfLocalsArea-(v->offset)));
                     break;

                   default:
                     bugT(t, "encode: unknown stdr tag for VAR");

                   }

                   emit("\tpushl\t%%eax");
                 }
               }
               break;

      case INT_CONST_TAG:
      case LONG_CONST_TAG:
      case UINT_CONST_TAG:
      case ULONG_CONST_TAG:

               /*
                *  The following cast is a bit sleezy, because there is
                *  actually four separate tree structs, one for each type
                *  of constant. However, the structs only differ in the
                *  type of the member holding the value. But on IA-32 these
                *  types are all the same so the cast is okay. Again, note
                *  we are assuming the compiler is executing on an IA-32!
                */
               i = ( IntConstTree ) t;
               if (i->expn.reg) break;

               /* mark tree as visited */
               i->expn.reg = setReg();

               emit("\tpushl\t$%d", i->value);
               break;

      case FP_CONST_TAG:
               fp = ( FpConstTree ) t;
               if (fp->expn.reg) break;

               /* mark tree as visited */
               fp->expn.reg = setReg();

               {
                 int lab = genLabel();
                 int size = computeSize(fp->expn.type);
                 char c = determineIA32Type(fp->expn.type);
                 char *typeStr;

                 switch(typeQuery(fp->expn.type)) {
            
                 case TYFLOAT:
                   typeStr = "float";
                   break;

                 case TYDOUBLE:
                   typeStr = "double";
                   break;

                 case TYLONGDOUBLE:
                   typeStr = "tfloat";
                   break;

                 default:
                   bugT(t, "encode: unexpected type in FP_CONST_TAG");
                 }

                 /* put constant value in memory in data section */
                 emit("\t.data");
                 emit("FPC%d:\t.%s %s", lab, typeStr, fp->value);
                 emit("\t.text");

                 /* load to fp register */
                 emit("\tfld%c\tFPC%d", c, lab);

                 /* alloc spot on top of stack */
                 emit("\tsubl\t$%d, %%esp", size);

                 /* put value on top of stack (and pop fp register stack) */
                 emit("\tfstp%c\t(%%esp)", c);

               }
               break;

      case STRING_LIT_TAG:
               s = ( StringLitTree ) t;
               if (s->expn.reg) break;

               /* mark tree as visited */
               s->expn.reg = setReg();

               emit("\t.data");
               labNum = genLabel();
               emit("L$%d:\t.asciz\t%s", labNum, s->string);
               emit("\t.text");
               emit("\tpushl\t$L$%d", labNum);
               break;

      case LABEL_TAG:
               la = ( LabelTree ) t;
               encodeLabel(la);
               break;

      case CASE_TAG:
               ca = ( CaseTree ) t;
               encodeCase(ca);
               break;

      case DEFAULT_TAG:
               d = ( DefaultTree ) t;
               encodeDefault(d);
               break;

      case SEQ_TAG:
               se = ( SeqTree ) t;
               encode( se->left );
               freeExpressionStatementRegister( se->left );
               encode( se->right );
               freeExpressionStatementRegister( se->right );
               break;

      case IFELSE_TAG:
               ie = ( IfElseTree ) t;
               encodeIfElse(ie);
               break;

      case SWITCH_TAG:
               sw = ( SwitchTree ) t;
               encodeSwitch(sw);
               break;

      case WHILE_TAG:
               w = ( WhileTree ) t;
               encodeWhile(w);
               break;

      case DOWHILE_TAG:
               dw = ( DoWhileTree ) t;
               encodeDoWhile(dw);
               break;

      case FOR_TAG:
               fo = ( ForTree ) t;
               encodeFor( fo );  
               break;

      case GOTO_TAG:
               g = ( GotoTree ) t;
               encodeGoto(g);
               break;

      case CONTINUE_TAG:
               co = ( ContinueTree ) t;
               encodeContinue(co);
               break;

      case BREAK_TAG:
               br = ( BreakTree ) t;
               encodeBreak(br);
               break;

      case RETURN_TAG:
               r = ( ReturnTree ) t;
               encode( ( Tree ) r->expr );
               encodeReturn(r);
               break;

      case ERROR_TAG:
               break;

      default:
             bugT(t,  "Invalid tree tag in encode()" );

   }

   return;
 }

/*
 *  Used to round frame sizes up to even multiple of 4.
 *
 */
static int roundEven4(int val)
{
  if ((val & 0x3) == 0) return val;

  return (val | 0x3) + 1;
}

/*
 *  encodeFunctionEntry: generates code for function entry.
 *
 *  We don't use the savedRegsSize and argBuildSize arguments.
 *
 *  The isStatic parameter tracks whether the function was declared
 *  with "static" storage class.
 *
 */
void encodeFunctionEntry(ST_ID name, int localsSize, int savedRegsSize,
                         int argBuildSize, BOOLEAN isStatic)
{
  int frameSize;

  /* quiet warnings */
  savedRegsSize = 0;
  argBuildSize = 0;

  frameSize = roundEven4(localsSize);

  /* need to remember this in order to generate code for LDECLS later */
  sizeOfLocalsArea = frameSize;

  emit("\t.text");
  if (!isStatic)
  {
    emit("\t.globl\t%s", st_get_id_str(name));
  }
  emit("%s:", st_get_id_str(name));
  emit("\tpushl\t%%ebp");
  emit("\tmovl\t%%esp, %%ebp");
  if (frameSize != 0)
    emit("\tsubl\t$%d, %%esp", frameSize);
}

/*
 *  encodeFunctionExit: generates code for function exit.
 *
 *  We don't use the savedRegsSize and argBuildSize arguments.
 *
 *  The isStatic parameter tracks whether the function was declared
 *  with "static" storage class.
 *
 */
void encodeFunctionExit(ST_ID name, int localsSize, int savedRegsSize,
                        int argBuildSize)
{
  int frameSize;

  /* quiet warnings */
  savedRegsSize = 0;
  argBuildSize = 0;

  frameSize = roundEven4(localsSize);

  emit("%s$exit:", st_get_id_str(name));
  if (frameSize != 0)
    emit("\taddl\t$%d, %%esp", frameSize);
  emit("\tpopl\t%%ebp");
  emit("\tret");
  emit("\t.data");
}

/*
 *  encodeConvert
 *
 *  This routine exploits the fact that we store all integer types in
 *  32 bits on the stack.
 *
 *  So the conversion possibilities are greatly restricted. The
 *  four possibilities are then 32-bit integer, 32-bit f.p., 64-bit f.p.
 *  and 80-bit f.p.
 *
 *  And the routine also exploits the fact that in the IA-32 all f.p.
 *  values are stored in a f.p. register using the 80-bit format. So
 *  basically a conversion is just a load of the value into a f.p. register
 *  and then store it back in the required format.
 *
 *  The complication is that, when converting from f.p. type to integer type,
 *  we need to manipulate the f.p. control word in order to set the rounding
 *  mode to "truncate" otherwise the "convert to integer" instruction will
 *  round, rather than truncate.
 *
 *  Also problem when going from unsigned integer types to f.p. types. Need
 *  to load "long long" into the f.p. register in order to make sure the
 *  sign bit is clear.
 *
 */
static void encodeConvert(UnopTree u)
{
  TypeTag destType;
  TypeTag sourceType;

  /* first some structural sanity checks */
  if (!u)
  {
    bug("null pointer passed to encodeConvert");
  }
  if (!u->left)
  {
    bug("null child pointer passed to encodeConvert");
  }

  /* mark tree as visited */
  u->expn.reg = setReg();

  destType = typeQuery(u->expn.type);

  /* if error, do nothing */
  if (destType == TYERROR) return;

  /* if destination type is void, just pop the stack */
  if (destType == TYVOID)
  {
    freeExpressionStatementRegister((Tree)(u->left));
    return;
  }

  /* weed out bug cases */
  switch (destType) {
  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
    break;

  case TYPOINTER:
    /* nothing to be done for pointer conversions */
    return;
  case TYVOID:
    bugT((Tree) u, "void destination type mishandled in encodeConvert()");
    break;
  case TYERROR:
    bugT((Tree) u, "error type seen in encodeConvert()");
    break;
  case TYSTRUCT:
    bugT((Tree) u, "struct type seen in encodeConvert()");
    break;
  case TYUNION:
    bugT((Tree) u, "union type seen in encodeConvert()");
    break;
  case TYARRAY:
    bugT((Tree) u, "array type seen in encodeConvert()");
    break;
  case TYFUNCTION:
    bugT((Tree) u, "function type seen in encodeConvert()");
    break;
  default:
    bugT((Tree) u, "unknown type seen in encodeConvert()");
  }

  sourceType = typeQuery(u->left->type);

  switch (sourceType) {
  case TYFLOAT:
  case TYDOUBLE:
  case TYLONGDOUBLE:

    switch (destType) {

      case TYFLOAT:
      case TYDOUBLE:
      case TYLONGDOUBLE:

        if (sourceType != destType)
        {
          /* load value to be converted into fp reg */
          emit("\tfld%c\t(%%esp)", determineIA32Type(u->left->type));

          /* free top of stack */
          emit("\taddl\t$%d, %%esp", computeSize(u->left->type));

          /* alloc space on top of stack for converted value */
          emit("\tsubl\t$%d, %%esp", computeSize(u->expn.type));

          /* put value on top of stack (and pop fp register stack) */
          emit("\tfstp%c\t(%%esp)", determineIA32Type(u->expn.type));
        }
        return;

      default:

        /* dest is some integer type */

        /* load value to be converted into fp reg */
        emit("\tfld%c\t(%%esp)", determineIA32Type(u->left->type));

        /* free top of stack */
        emit("\taddl\t$%d, %%esp", computeSize(u->left->type));

        /* set rounding mode to be "truncate" */
        emit("\tsubl\t$4, %%esp");     /* alloc two words on top of stack */
        emit("\tfnstcw\t(%%esp)");     /* store fp control word in first word */
        emit("\tmovw\t(%%esp), %%ax"); /* put control word in %ax */
        emit("\tmovb\t$12, %%ah");     /* set RC bits to "truncate" */
        emit("\tmovw\t%%ax, 2(%%esp)");/* put modified cntrl word in 2nd slot */
        emit("\tfldcw\t2(%%esp)");     /* load modified control word */
        emit("\tmovw\t(%%esp), %%ax"); /* put old control word in %ax */
        emit("\taddl\t$4, %%esp");     /* free the two words */

        /* alloc space for integer result on top of stack */
        emit("\tsubl\t$4, %%esp");

        /* convert value to signed int */
        emit("\tfistpl\t(%%esp)");

        /* re-establish old fp control word */
        emit("\tsubl\t$2, %%esp");      /* alloc one word on top of stack */
        emit("\tmovw\t%%ax, (%%esp)");  /* put original cntrl word on stack */
        emit("\tfldcw\t(%%esp)");       /* re-load original control word */
        emit("\taddl\t$2, %%esp");      /* free word on top of stack */

        /* unsigned destination types need the upper bytes cleared */
        if (destType == TYUNSIGNEDSHORTINT)
        {
          emit("\tmovw\t$0, 2(%%esp)");
        }
        else if (destType == TYUNSIGNEDCHAR)
        {
          emit("\tmovw\t$0, 2(%%esp)");
          emit("\tmovb\t$0, 1(%%esp)");
        }

        /* all done */
        return;
    }

  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:

    switch (destType) {

      case TYFLOAT:
      case TYDOUBLE:
      case TYLONGDOUBLE:

        /* if unsigned then load 64-bit integer into f.p. register */
        if (sourceType == TYUNSIGNEDLONGINT || sourceType == TYUNSIGNEDCHAR ||
            sourceType == TYUNSIGNEDINT || sourceType == TYUNSIGNEDSHORTINT)
        {
          /* pop low-order bytes into eax */
          emit("\tpopl\t%%eax");

          /* zero out high-order bytes */
          emit("\tpushl\t$0");

          /* push back the low-order bytes */
          emit("\tpushl\t%%eax");

          /* load the zero-extended value to a fp reg */
          emit("\tfildll\t(%%esp)");

          /* free the top of stack */
          emit("\taddl\t$8, %%esp");
        }
        else
        {
          /* just must move the value to a fp reg */
          emit("\tfildl\t(%%esp)");

          /* free the top of stack */
          emit("\taddl\t$4, %%esp");
        }

        /* alloc space on top of stack for converted value */
        emit("\tsubl\t$%d, %%esp", computeSize(u->expn.type));

        /* put value on top of stack (and pop fp register stack) */
        emit("\tfstp%c\t(%%esp)", determineIA32Type(u->expn.type));

        /* all done */
        return;

      /*  destination and source are both integer types:
       *    if destination is unsigned, clear the upper bytes
       */
      default:
        if (destType == TYUNSIGNEDSHORTINT)
        {
          emit("\tmovw\t$0, 2(%%esp)");
        }
        else if (destType == TYUNSIGNEDCHAR)
        {
          emit("\tmovw\t$0, 2(%%esp)");
          emit("\tmovb\t$0, 1(%%esp)");
        }
        return;
    }

  case TYPOINTER:
    /* nothing to be done but should never reach here */
    return;
  case TYVOID:
    bugT((Tree) u, "void source type seen in encodeConvert()");
    break;
  case TYERROR:
    bugT((Tree) u, "error type seen in encodeConvert()");
    break;
  case TYSTRUCT:
    bugT((Tree) u, "struct type seen in encodeConvert()");
    break;
  case TYUNION:
    bugT((Tree) u, "union type seen in encodeConvert()");
    break;
  case TYARRAY:
    bugT((Tree) u, "array type seen in encodeConvert()");
    break;
  case TYFUNCTION:
    bugT((Tree) u, "function type seen in encodeConvert()");
    break;
  default:
    bugT((Tree) u, "unknown type seen in encodeConvert()");
  }
}

/*
 *  Arrays and functions are always passed by their base address.
 *
 *  Also all parameters are at least 4 bytes in size. This is
 *  because we always keep the run-time stack aligned on 4 byte boundaries.
 */
int computeParamSize(Type t)
{
  switch (typeQuery(t)) {

  case TYARRAY:
  case TYFUNCTION:
    return 4;

  default:
  {
    int tmp;
    tmp = computeSize(t);
    if (tmp < 4) tmp = 4;
    return tmp;
  }

  }
  return 0; /* NOTREACHED */
}

/*
 *  Given address of stuct/union on top of stack, replace address
 *  with a copy of the struct/union.
 */
static void copyStructUnionToTopOfStack(int size)
{
  emit("\tpopl\t%%eax");             /* pop base address */
  emit("\tsubl\t$%d, %%esp", size);  /* alloc space for struct/union */
  emit("\tmovl\t%%esp, %%edx");      /* remember where to copy to */
  emit("\tpushl\t$%d", size);        /* push length of struct/union */
  emit("\tpushl\t%%eax");            /* push base address */
  emit("\tpushl\t%%edx");            /* push destination of copy */
  emit("\tcall\tmemcpy");            /* call memcpy */
  emit("\taddl\t$12, %%esp");        /* pop memcpy args */
}

/*
 *  Recursive function to traverse and encode FCALL parameter list in reverse
 *  order.
 *
 *  Need to track the amount of stack allocated in order to pop the
 *  arguments after the call.
 *
 *  Don't need to worry about alignment here because everything is a multiple
 *  of four bytes. (Remember that structs are padded out to worst case
 *  alignment.)
 *
 *  But do need to worry about arguments of char or short type. Need to
 *  remember to round up the space allocated for these to 4.
 *
 *  The second parameter is the function return type. If the return type
 *  is a struct/union then the size of the first argument is 4: a pointer
 *  to where to put the return value.
 *
 *  Special care must be taken with struct args. Encoding arg will only
 *  produce the base address on top of the stack. I must pop the base
 *  address and copy the struct on to the top of stack.
 *
 */
static int encodeFcallHelper(ExpTree e, Type retType, BOOLEAN isRoot)
{
  int ret;

  if (((Tree) e) -> tag == ESEQ_TAG)
  {
    ret = encodeFcallHelper(((EseqTree)e) -> right, retType, FALSE);

    if (isRoot)
    {
      if (isStructUnionType(retType))
      {
        ret += 4;
        encode((Tree)(((EseqTree)e) -> left));
      }
      else
      {
        int tmp;
        tmp = computeParamSize(((EseqTree)e) -> left -> type);
        if (tmp < 4) tmp = 4;
        ret += tmp;
        encode((Tree)(((EseqTree)e) -> left));
        if (isStructUnion(((EseqTree)e) -> left))
        {
          copyStructUnionToTopOfStack(computeSize(((EseqTree)e)->left->type));
        }
      }
    }
    else
    {
      int tmp;
      tmp = computeParamSize(((EseqTree)e) -> left -> type);
      if (tmp < 4) tmp = 4;
      ret += tmp;
      encode((Tree)(((EseqTree)e) -> left));
      if (isStructUnion(((EseqTree)e) -> left))
      {
        copyStructUnionToTopOfStack(computeSize(((EseqTree)e)->left->type));
      }
    }
      
  }
  else
  {
    if (isRoot)
    {
      if (isStructUnionType(retType))
      {
        ret = 4;
        encode((Tree) e);
      }
      else
      {
        int tmp;
        tmp = computeParamSize(e -> type);
        if (tmp < 4) tmp = 4;
        ret = tmp;
        encode((Tree) e);
        if (isStructUnion(e))
        {
          copyStructUnionToTopOfStack(computeSize(e->type));
        }
      }
    }
    else
    {
      int tmp;
      tmp = computeParamSize(e -> type);
      if (tmp < 4) tmp = 4;
      ret = tmp;
      encode((Tree) e);
      if (isStructUnion(e))
      {
        copyStructUnionToTopOfStack(computeSize(e->type));
      }
    }
  }

  return ret;
}

/*
 *  Encode a function call.
 *
 *  Just need to process the argument list in reverse order to get
 *  arguments evaluated on top of the stack. Then emit the call instruction.
 *  When function returns need to push the result on top of stack.
 */
static void encodeFcall(FcallTree b)
{
  int argsSpace;

  argsSpace = 0;

  /* Use recursive helper function to process arguments in reverse order. */
  if (b->params != NULL)
    argsSpace = encodeFcallHelper(b->params,  b->expn.type, TRUE);

  /* What are we calling? */
  encode((Tree) b->func);

  /* Do the call */
  emit("\tpopl\t%%eax");
  emit("\tcall\t*%%eax");

  /* Remove the args */
  if (argsSpace != 0)
    emit("\taddl\t$%d, %%esp", argsSpace);

  /* Mark the tree as visited */
  b->expn.reg = setReg();
 
  /* If there is a return value, push it onto stack */
  switch (typeQuery(((ExpTree) b)-> type))
  {
    case TYVOID:        /* Do Nothing */
      break;

    case TYFLOAT:
    case TYDOUBLE:
    case TYLONGDOUBLE:  /* Floating point */

      /* alloc space on top of stack for return value */
      emit("\tsubl\t$%d, %%esp", computeSize(b->expn.type));

      /* put value on top of stack (and pop fp register stack) */
      emit("\tfstp%c\t(%%esp)", determineIA32Type(b->expn.type));

      break;

    case TYSIGNEDLONGINT:
    case TYSIGNEDSHORTINT:
    case TYSIGNEDINT: 
    case TYUNSIGNEDLONGINT:
    case TYUNSIGNEDSHORTINT:
    case TYUNSIGNEDINT:
    case TYUNSIGNEDCHAR:
    case TYSIGNEDCHAR:
    case TYSTRUCT:
    case TYUNION:
    case TYARRAY:
    case TYPOINTER:         /* Integer or Pointer */

      /* push return value (in %eax) onto stack */
      emit("\tpushl\t%%eax");

      break;

    default:
      bug("unexpected return type");
  }

}

/*
 *  This is useful on some architectures but not for our approach to IA-32.
 */
void encodeMoveParametersToLocalSlots(ParamList list, Tree t)
{
  /* do nothing */

  /* quiet warnings */
  list = NULL;
  t = NULL;
}

/*
 *  struct/union done via memcpy.
 */
static void encodeStructUnionAssign(BinopTree b)
{
  int structUnionSize;

  structUnionSize = computeSize( b->expn.type );

  /*
   *  The stack now contains lhs (dest) and rhs (src) with rhs on top.
   *  I need to have len, rhs (src) and lhs (dest) with dest on top.
   *
   *  Tricky: I need to leave the src address on top of the stack as
   *  the result of the assignment.
   */
  emit("\tpopl\t%%eax");                 /* pop rhs (src)  */
  emit("\tpopl\t%%edx");                 /* pop lhs (dest) */
  emit("\tpushl\t%%eax");                /* push src       */
  emit("\tpushl\t$%d", structUnionSize); /* push len       */
  emit("\tpushl\t%%eax");                /* push src       */
  emit("\tpushl\t%%edx");                /* push dest      */

  /* now call memcpy */
  emit( "\tcall\tmemcpy" );

  /* pop memcpy's args */
  emit("\taddl\t$12, %%esp");

  /* mark the tree as visited by the code generator */
  b->expn.reg = setReg();
}

static void encodeAssign(BinopTree b)
{
  /* special handling needed for structs and unions */
  if (isStructUnionType(b->expn.type))
  {
    encodeStructUnionAssign(b);
  }
  else
  {
    encodeAssignDetail(b->expn.type, (Tree)b);
  }
}

static void encodeReturn(ReturnTree r)
{
  if ( r->expr == NULL )
  {
     emit( "\tjmp\t%s$exit", st_get_id_str( gCurrentFunctionName ) );
  }
  else
  {
     switch (typeQuery(r->expr->type))
     {
     case TYSIGNEDLONGINT:
     case TYSIGNEDSHORTINT:
     case TYSIGNEDINT:
     case TYUNSIGNEDLONGINT:
     case TYUNSIGNEDSHORTINT:
     case TYUNSIGNEDINT:
     case TYUNSIGNEDCHAR:
     case TYSIGNEDCHAR:
     case TYPOINTER:
        emit("\tpopl\t%%eax");
        break;

     case TYFLOAT:
     case TYDOUBLE:
     case TYLONGDOUBLE:
        emit("\tfld%c\t(%%esp)", determineIA32Type(r->expr->type));
        emit("\taddl\t$%d, %%esp", computeSize(r->expr->type));
        break;

     case TYSTRUCT:
     case TYUNION:
        /* memcpy the struct content to where the extra argument points */	

        /* source address is now on top of the stack: pop it into %eax */
        emit("\tpopl\t%%eax");

        /* push the struct size */
        emit("\tpushl\t$%d", computeSize( r->expr->type ) );

        /* push the source address */
        emit("\tpushl\t%%eax");

        /* push destination address, which is in first locals slot */
        emit("\tmovl\t8(%%ebp), %%eax");
        emit("\tpushl\t%%eax");

        /* do memory copy*/
        emit("\tcall\tmemcpy");

        /* pop memcpy's args */
        emit("\taddl\t$12, %%esp");

        /* return the struct address */
        emit("\tmovl\t8(%%ebp), %%eax");

        break;

     case TYERROR:
       break; /* should never get here? */

     case TYVOID:
       bugT((Tree) r, "void type in encodeReturn");

     case TYARRAY:
       bugT((Tree) r, "array type in encodeReturn");

     case TYFUNCTION:
       bugT((Tree) r, "function type in encodeReturn");

     default:
       bugT((Tree) r, "unknown type in encodeReturn");
     }

     emit( "\tjmp\t%s$exit", st_get_id_str( gCurrentFunctionName ) );

  }

}

/*
 *  We are not using registers but we need to pop the stack.
 */
void freeExpressionStatementRegister(Tree t)
{
  if (!isExpTree(t)) return;

  switch(typeQuery(((ExpTree)t)->type)) {

  case TYFLOAT:
  case TYSIGNEDLONGINT:
  case TYSIGNEDSHORTINT:
  case TYSIGNEDINT:
  case TYUNSIGNEDLONGINT:
  case TYUNSIGNEDSHORTINT:
  case TYUNSIGNEDINT:
  case TYUNSIGNEDCHAR:
  case TYSIGNEDCHAR:
  case TYPOINTER:
  case TYSTRUCT:
  case TYUNION:
    emit("\taddl\t$4, %%esp");
    break;

  case TYDOUBLE:
    emit("\taddl\t$8, %%esp");
    break;

  case TYLONGDOUBLE:
    emit("\taddl\t$12, %%esp");
    break;

  case TYERROR:
  case TYVOID:
    break;

  case TYARRAY:
  case TYFUNCTION:
    bugT(t, "unexpected type in freeExpressionStatementRegister");

  default:
    bugT(t, "unknown type in freeExpressionStatementRegister");
  }
}

/*
 *  Need to adjust running offset as shift from params to locals?
 *
 *  On IA-32 I want to reset the offset to zero since params are positive
 *  offsets from %ebp and locals are negative offsets from %ebp.
 */
int adjustFunctionOffset(int currentOffset)
{
  /* quiet warning */
  currentOffset = 0;

  return 0;
}


